import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import AsyncExitStack
from functools import reduce
from typing import (
    AsyncIterable,
    Awaitable,
    ClassVar,
    Generic,
    Literal,
    Protocol,
    TypeVar,
    Union,
    final,
    get_origin,
    get_type_hints,
    runtime_checkable,
)

import anyio
import asyncstdlib
import pydantic
from aiostream import pipe, stream

from nova import api
from nova.actions import Action, MovementController
from nova.core import logger
from nova.core.movement_controller import movement_to_motion_state
from nova.types import MotionState, MovementResponse, Pose, RobotState


class RobotCellError(Exception):
    """Base exception for all robot cell specific error"""


class RobotMotionError(RobotCellError):
    """Robot can not move as requested"""


class RobotCellKeyError(KeyError):
    pass


class ConfigurablePeriphery:
    """A device which is configurable"""

    all_classes: ClassVar[dict] = {}

    def __init_subclass__(cls, is_abstract=False):
        super().__init_subclass__()
        if not is_abstract:
            assert (
                hasattr(cls, "Configuration")
                and get_origin(get_type_hints(cls.Configuration)["type"]) is Literal
            ), f"{cls.__name__} has no type literal"
            assert ConfigurablePeriphery.Configuration is not cls.Configuration
            cls.all_classes[cls.Configuration] = cls

    class Configuration(pydantic.BaseModel):
        """Minimum configuration of a configurable periphery

        Args:
            id: A unique id to reference the periphery
        """

        model_config = pydantic.ConfigDict(frozen=True)

        type: str
        id: str

    _configuration: Configuration

    def __init__(self, configuration: Configuration, **kwargs):
        super().__init__(**kwargs)
        self._configuration = configuration

    @property
    def configuration(self):
        return self._configuration

    @property
    def id(self):
        return self.configuration.id


class Device(ABC):
    """A device that takes care of lifetime management"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._is_active = False

    async def open(self) -> None:
        """Allocates the external hardware resource (i.e. establish a connection)"""
        self._is_active = True

    async def close(self):
        """Release the external hardware (i.e. close connection or set mode of external hardware back)"""
        self._is_active = False

    async def restart(self):
        if self._is_active:
            await self.close()
        await self.open()

    @final
    async def __aenter__(self):
        await self.open()
        return self

    @final
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


T = TypeVar("T")


class AsyncCallableDevice(Generic[T], Device):
    """An awaitable external function or service in the robot cell"""

    async def __call__(self, *args, **kwargs) -> Awaitable[T]:
        if not self._is_active:
            raise ValueError("The device is not activated.")
        return await self._call(*args)

    @abstractmethod
    async def _call(self, key, *args) -> Awaitable[T]:
        """The implementation of the call method. AbstractAwaitable guarantees that the device is activated.

        Args:
            key: A key that represents the id of the external function or service that is called
            *args: Parameters of the external callable

        Returns: the returned values of the external called function or service
        """


ValueType = Union[int, str, bool, float, Pose]


@runtime_checkable
class InputDevice(Protocol):
    """A device which supports reading from"""

    async def read(self, key: str) -> ValueType:
        """Read a value given its key"""


@runtime_checkable
class OutputDevice(Protocol):
    """A device which supports writing to"""

    async def write(self, key: str, value: ValueType) -> None:
        """Write a value given its key and the new value"""


@runtime_checkable
class IODevice(InputDevice, OutputDevice, Protocol):
    """A device which supports reading and writing"""


class AbstractDeviceState(Protocol):
    """A state of a device"""

    def __eq__(self, other: "AbstractDeviceState") -> bool:
        """Check if the state is equal to another state"""


@runtime_checkable
class StoppableDevice(Protocol):
    """A device that can be stopped"""

    async def stop(self) -> None:
        """Stop the device"""


@runtime_checkable
class StateStreamingDevice(Protocol):
    """A device which supports streaming its state"""

    def stream_state(self, rate_msecs: int) -> AsyncIterable[AbstractDeviceState]:
        """Read a value given its key
        Args:
            rate: The rate at which the state should be streamed
        """


class AbstractRobot(Device):
    """An interface for real and simulated robots"""

    _id: str

    def __init__(self, id: str, **kwargs):
        super().__init__(**kwargs)
        self._id = id

    @property
    def id(self):
        return self._id

    @abstractmethod
    async def _plan(
        self,
        actions: list[Action],
        tcp: str,
        start_joint_position: tuple[float, ...] | None = None,
        optimizer_setup: api.models.OptimizerSetup | None = None,
    ) -> api.models.JointTrajectory:
        """Plan a trajectory for the given actions

        Args:
            actions (list[Action] | Action): The actions to be planned. Can be a single action or a list of actions.
                Only motion actions are considered for planning.
            tcp (str): The id of the tool center point (TCP)
            start_joint_position (tuple[float, ...] | None): The starting joint position. If None, the current joint

        Returns:
            api.models.JointTrajectory: The planned joint trajectory
        """

    async def plan(
        self,
        actions: list[Action] | Action,
        tcp: str,
        start_joint_position: tuple[float, ...] | None = None,
        optimizer_setup: api.models.OptimizerSetup | None = None,
    ) -> api.models.JointTrajectory:
        """Plan a trajectory for the given actions

        Args:
            actions (list[Action] | Action): The actions to be planned. Can be a single action or a list of actions.
                Only motion actions are considered for planning.
            tcp (str): The id of the tool center point (TCP)
            start_joint_position: the initial position of the robot
            start_joint_position (tuple[float, ...] | None): The starting joint position. If None, the current joint
            optimizer_setup (api.models.OptimizerSetup | None): The optimizer setup to be used for planning

        Returns:
            api.models.JointTrajectory: The planned joint trajectory
        """
        if not isinstance(actions, list):
            actions = [actions]

        if len(actions) == 0:
            raise ValueError("No actions provided")

        return await self._plan(
            actions=actions,
            tcp=tcp,
            start_joint_position=start_joint_position,
            optimizer_setup=optimizer_setup,
        )

    @abstractmethod
    def _execute(
        self,
        joint_trajectory: api.models.JointTrajectory,
        tcp: str,
        actions: list[Action],
        movement_controller: MovementController | None,
    ) -> AsyncIterable[MovementResponse]:
        """Execute a planned motion

        Args:
            joint_trajectory (api.models.JointTrajectory): The planned joint trajectory
            tcp (str): The id of the tool center point (TCP)
            actions (list[Action] | Action | None): The actions to be executed. Defaults to None.
            movement_controller (MovementController): The movement controller to be used. Defaults to move_forward
        """

    async def stream_execute(
        self,
        joint_trajectory: api.models.JointTrajectory,
        tcp: str,
        actions: list[Action] | Action,
        movement_controller: MovementController | None = None,
    ) -> AsyncIterable[MotionState]:
        """Execute a planned motion

        Args:
            joint_trajectory (api.models.JointTrajectory): The planned joint trajectory
            tcp (str): The id of the tool center point (TCP)
            actions (list[Action] | Action | None): The actions to be executed. Defaults to None.
            movement_controller (MovementController): The movement controller to be used. Defaults to move_forward
        """
        if not isinstance(actions, list):
            actions = [actions]

        def is_movement(movement_response: MovementResponse) -> bool:
            return any(
                (
                    isinstance(movement_response, api.models.ExecuteTrajectoryResponse)
                    and isinstance(movement_response.actual_instance, api.models.Movement),
                    isinstance(movement_response, api.models.StreamMoveResponse),
                )
            )

        def movement_response_to_motion_state(
            movement_response: MovementResponse, *_
        ) -> MotionState:
            if isinstance(movement_response, api.models.ExecuteTrajectoryResponse):
                return movement_to_motion_state(movement_response.actual_instance)
            if isinstance(movement_response, api.models.StreamMoveResponse):
                return movement_to_motion_state(movement_response)
            assert False, f"Unexpected movement response: {movement_response}"

        execute_response_stream = self._execute(
            joint_trajectory, tcp, actions, movement_controller=movement_controller
        )
        motion_states = (
            stream.iterate(execute_response_stream)
            | pipe.filter(is_movement)
            | pipe.map(movement_response_to_motion_state)
        )

        async with motion_states.stream() as motion_states_stream:
            async for motion_state in motion_states_stream:
                yield motion_state

    async def execute(
        self,
        joint_trajectory: api.models.JointTrajectory,
        tcp: str,
        actions: list[Action] | Action,
        movement_controller: MovementController | None = None,
    ) -> None:
        """Execute a planned motion

        Args:
            joint_trajectory (api.models.JointTrajectory): The planned joint trajectory
            tcp (str): The id of the tool center point (TCP)
            actions (list[Action] | Action): The actions to be executed.
            movement_controller (MovementController): The movement controller to be used. Defaults to move_forward
        """
        async for _ in self.stream_execute(
            joint_trajectory, tcp, actions, movement_controller=movement_controller
        ):
            pass

    async def stream_plan_and_execute(
        self,
        actions: list[Action] | Action,
        tcp: str,
        start_joint_position: tuple[float, ...] | None = None,
    ) -> AsyncIterable[MotionState]:
        joint_trajectory = await self.plan(actions, tcp, start_joint_position=start_joint_position)
        async for motion_state in self.stream_execute(joint_trajectory, tcp, actions):
            yield motion_state

    async def plan_and_execute(
        self,
        actions: list[Action] | Action,
        tcp: str,
        start_joint_position: tuple[float, ...] | None = None,
    ) -> None:
        joint_trajectory = await self.plan(actions, tcp, start_joint_position=start_joint_position)
        await self.execute(joint_trajectory, tcp, actions, movement_controller=None)

    @abstractmethod
    async def get_state(self, tcp: str | None = None) -> RobotState:
        """Current state (pose, joints) of the robot based on the tcp.

        Args:
            tcp (str): The id of the tool center point (TCP) to be used for tcp_pose in response. If not set,
                the flange pose is returned as tcp_pose.

        Returns: the current state of the robot

        """

    @abstractmethod
    async def joints(self) -> tuple:
        """Return the current joint positions of the robot

        Returns: the current joint positions

        """

    @abstractmethod
    async def tcp_pose(self, tcp: str | None = None) -> Pose:
        """Return the current pose of the robot based on the tcp

        Args:
            tcp (str): The id of the tool center point (TCP) to be used for tcp_pose in response. If not set,
                the flange pose is returned as tcp_pose.

        Returns: the current pose of the robot

        """

    @abstractmethod
    async def tcps(self) -> list[api.models.RobotTcp]:
        """Return all TCPs that are configured on the robot with corresponding offset from flange as pose

        Returns: the TCPs of the robot

        """

    @abstractmethod
    async def tcp_names(self) -> list[str]:
        """Return the name of all TCPs that are configured on the robot

        Returns: a list of all TCPs

        """

    @abstractmethod
    async def active_tcp(self) -> api.models.RobotTcp:
        """Return the active TCP of the robot

        Returns: the active TCP

        """

    @abstractmethod
    async def active_tcp_name(self) -> str:
        """Return the name of the active TCP of the robot

        Returns: the name of the active TCP

        """

    @abstractmethod
    async def stop(self):
        """Stop behaviour of the robot"""


class AbstractController(Device):
    @abstractmethod
    def get_robots(self) -> dict[str, AbstractRobot]:
        """Return all robots that are connected to the controller

        Returns: a dict with {robot_name: robot}
        """


class AbstractTimer(Device):
    """A timer"""

    @abstractmethod
    async def __call__(self, duration: float) -> None:
        """Wait for a duration ms"""


class Timer(ConfigurablePeriphery, AbstractTimer):
    """A real timer (blocking the execution)"""

    class Configuration(ConfigurablePeriphery.Configuration):
        type: Literal["timer", "simulated_timer"] = "timer"
        id: str = "timer"

    def __init__(self, configuration: Configuration = Configuration()):
        super().__init__(configuration)

    async def __call__(self, duration: float):
        await asyncio.sleep(duration / 1000)


class RobotCell:
    """Access a simulated or real robot"""

    _devices: dict

    def __init__(self, timer: AbstractTimer | None = None, **kwargs):
        if timer is None:
            timer = Timer()
        devices = {"timer": timer, **kwargs}
        # TODO: if "timer" has not the same id it cannot correctly be serialized/deserialized currently
        for device_name, device in devices.items():
            if device and device_name != device.id:
                raise ValueError(
                    f"The device name should match its name in the robotcell but are '{device_name}' and '{device.id}'"
                )
        self._devices = devices
        self._device_exit_stack = AsyncExitStack()

    @property
    def devices(self) -> dict:
        return self._devices

    def set_configurations(self, configurations: list[ConfigurablePeriphery.Configuration]):
        """Set the configurations of all devices that are attached to the robot cell

        Args:
            configurations: the configurations of the robot cell

        """
        self._devices.clear()
        self.apply_configurations(configurations)

    def apply_configurations(self, configurations: list[ConfigurablePeriphery.Configuration]):
        """Applies all given device configurations to the robot cell. If the id is already in the
        robot cell the device gets overriden.

        Args:
            configurations: the given device configurations

        """
        for configuration in configurations:
            logger.info(f"Setup device with configuration: {configuration}...")
            device_id = configuration.id
            result = ConfigurablePeriphery.all_classes[type(configuration)](
                configuration=configuration
            )
            self._devices[device_id] = result

    def to_configurations(self) -> list[ConfigurablePeriphery.Configuration]:
        """Return the configurations of all devices that are attached to the robot cell

        Returns:
            [list[ConfigurablePeriphery.Configuration]]: the configurations of the robot cell

        """
        # TODO: remove 'hasattr(device, "configuration")' See: https://wandelbots.atlassian.net/browse/WP-554
        return [
            device.configuration
            for device in self._devices.values()
            if hasattr(device, "configuration")
        ]

    @classmethod
    def from_configurations(cls, configurations: list[ConfigurablePeriphery.Configuration]):
        """Construct a new robot_cell from the device configurations

        Returns:
            [RobotCell]: the newly created RobotCell

        """
        robot_cell = cls()
        robot_cell.apply_configurations(configurations)
        return robot_cell

    def get_controllers(self) -> list[AbstractController]:
        return list(
            controller
            for controller in self._devices.values()
            if isinstance(controller, AbstractController)
        )

    def get_controller(self, name: str) -> AbstractController:
        """Return the controller by its name"""
        controller = self[name]
        if not isinstance(controller, AbstractController):
            raise ValueError(f'Found no controller with name "{name}".')
        return controller

    def get_robots(self) -> dict[str, AbstractRobot]:
        controllers = self.get_controllers()
        return reduce(
            lambda a, b: a.update(b) or a,
            (controller.get_robots() for controller in controllers),
            {},
        )

    def get_robot_ids(self) -> list[str]:
        # TODO this should return a set
        return list(self.get_robots().keys())

    def get_robot(self, robot_id: str) -> AbstractRobot:
        return self.get_robots()[robot_id]

    @asyncstdlib.cached_property
    async def tcps(self) -> dict[str, set[str]]:
        """Return a mapping of all TCPs to the robots that have them configured"""
        result = defaultdict(set)
        for robot_name, robot in self.get_robots().items():
            tcp_names = await robot.tcp_names()
            for tcp_name in tcp_names:
                result[tcp_name].add(robot_name)
        return result

    @property
    def timer(self) -> AbstractTimer:
        return self["timer"]

    async def stop(self):
        """Stop the robot cell"""
        stoppable_devices = [
            device for device in self._devices.values() if isinstance(device, StoppableDevice)
        ]
        if not stoppable_devices:
            return

        async with anyio.create_task_group() as task_group:
            for device in stoppable_devices:
                task_group.start_soon(device.stop)

    async def stream_state(self, rate_msecs: int):
        """Stream the state of the robot cell"""
        state_streaming_devices = [
            device for device in self._devices.values() if isinstance(device, StateStreamingDevice)
        ]
        if not state_streaming_devices:
            return

        state_streams = [device.stream_state(rate_msecs) for device in state_streaming_devices]
        async with stream.merge(*state_streams).stream() as devices_state_stream:
            async for state in devices_state_stream:
                yield state

    # async def open(self):
    #    for device in self.values():
    #       await self._exit_stack.enter_async_context(device)
    #    await super().open()
    #    raise NotImplementedError()

    # async def close(self):
    #    raise NotImplementedError()

    async def __aenter__(self):
        for device in self._devices.values():
            await self._device_exit_stack.enter_async_context(device)
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        return await self._device_exit_stack.__aexit__(exc_type, exc_value, exc_traceback)

    def __getitem__(self, item):
        try:
            return self._devices.__getitem__(item)
        except KeyError as exc:
            raise RobotCellKeyError(item) from exc

    def __delitem__(self, key):
        try:
            return self._devices.__delitem__(key)
        except KeyError as exc:
            raise RobotCellKeyError(key) from exc



from pathlib import Path
from unittest.mock import patch

import pytest

from nova_rerun_bridge.trajectory import TimingMode, log_motion
from tests.fixtures.motion_loader import load_motion_data

FIXTURES_DIR = Path(__file__).parent.parent.parent / "tests" / "fixtures" / "recorded_motions"


@pytest.fixture
def motion_data():
    """Load recorded motion data."""
    fixture_path = FIXTURES_DIR / "example_motion.json"
    return load_motion_data(fixture_path)


@pytest.mark.asyncio
async def test_log_motion_with_recorded_data(motion_data):
    """Test log_motion using recorded data."""
    with patch("rerun.log") as mock_log:
        # Call log_motion with recorded data
        log_motion(
            motion_id=motion_data["motion_id"],
            model_from_controller=motion_data["model_from_controller"],
            motion_group=motion_data["motion_group"],
            optimizer_config=motion_data["optimizer_config"],
            trajectory=motion_data["trajectory"],
            collision_scenes=motion_data["collision_scenes"],
            time_offset=0.0,
            timing_mode=TimingMode.RESET,
        )

        # Verify rerun calls
        mock_log.assert_called()



from nova import actions, api, types
from nova.cell.cell import Cell
from nova.core.controller import Controller
from nova.core.logging import logger
from nova.core.motion_group import MotionGroup
from nova.core.movement_controller import speed_up as speed_up_movement_controller
from nova.core.nova import Nova
from nova.function import wrap as program
from nova.types import MotionSettings
from nova.version import version

__version__ = version


__all__ = [
    "Nova",
    "Cell",
    "MotionGroup",
    "Controller",
    "speed_up_movement_controller",
    "api",
    "types",
    "actions",
    "MotionSettings",
    "logger",
    "program",
    "__version__",
]



import nova.api as api
from nova.cell.controllers import virtual_controller
from nova.core.controller import Controller
from nova.core.exceptions import ControllerNotFound
from nova.core.gateway import ApiGateway
from nova.core.robot_cell import RobotCell

# This is the default value we use to wait for add_controller API call to complete.
DEFAULT_ADD_CONTROLLER_TIMEOUT = 25

# This is the default value we use when we wait for a controller to be ready.
DEFAULT_WAIT_FOR_READY_TIMEOUT = 25


class Cell:
    """A representation of a robot cell, providing high-level operations on controllers."""

    def __init__(self, api_gateway: ApiGateway, cell_id: str):
        """
        Initializes a Cell instance.
        Args:
            api_gateway (ApiGateway): The underlying gateway for making API calls.
            cell_id (str): The unique identifier for the cell.
        """
        self._api_gateway = api_gateway
        self._cell_id = cell_id

    @property
    def cell_id(self) -> str:
        """
        Returns unique identifier for this cell.
        Returns:
            str: The cell ID.
        """
        return self._cell_id

    def _create_controller(self, controller_id: str) -> Controller:
        return Controller(
            configuration=Controller.Configuration(
                nova_api=self._api_gateway.host,
                nova_access_token=self._api_gateway.access_token,
                nova_username=self._api_gateway.username,
                nova_password=self._api_gateway.password,
                cell_id=self._cell_id,
                controller_id=controller_id,
                id=controller_id,
            )
        )

    async def add_virtual_robot_controller(
        self,
        name: str,
        controller_type: api.models.VirtualControllerTypes,
        controller_manufacturer: api.models.Manufacturer,
        timeout: int = DEFAULT_ADD_CONTROLLER_TIMEOUT,
        wait_for_ready_timeout: int = DEFAULT_WAIT_FOR_READY_TIMEOUT,
        position: str | None = None,
    ) -> Controller:
        return await self.add_controller(
            robot_controller=virtual_controller(
                name=name,
                type=controller_type,
                manufacturer=controller_manufacturer,
                position=position,
            ),
            add_timeout=timeout,
            wait_for_ready_timeout=wait_for_ready_timeout,
        )

    async def ensure_virtual_robot_controller(
        self,
        name: str,
        controller_type: api.models.VirtualControllerTypes,
        controller_manufacturer: api.models.Manufacturer,
        timeout: int = DEFAULT_ADD_CONTROLLER_TIMEOUT,
        wait_for_ready_timeout: int = DEFAULT_WAIT_FOR_READY_TIMEOUT,
    ) -> Controller:
        return await self.ensure_controller(
            robot_controller=virtual_controller(
                name=name, type=controller_type, manufacturer=controller_manufacturer
            ),
            add_timeout=timeout,
            wait_for_ready_timeout=wait_for_ready_timeout,
        )

    async def add_controller(
        self,
        robot_controller: api.models.RobotController,
        add_timeout: int = DEFAULT_ADD_CONTROLLER_TIMEOUT,
        wait_for_ready_timeout: int = DEFAULT_WAIT_FOR_READY_TIMEOUT,
    ) -> Controller:
        """
        Add a robot controller to the cell and wait for it to get ready.
        Args:
            robot_controller (api.models.RobotController): The robot controller to add. You can use helper functions from nova to create these configs easily,
                see :func:`nova.cell.abb_controller`, :func:`nova.cell.fanuc_controller`, :func:`nova.cell.kuka_controller`,
                :func:`nova.cell.universal_robots_controller`, :func:`nova.cell.virtual_controller`, :func:`nova.cell.yaskawa_controller`.
            add_timeout (int): The time to wait for the controller to be added (default: 25).
            wait_for_ready_timeout (int): The time to wait for the controller to be ready (default: 25).

        Returns:
            Controller: The added Controller object.
        """
        await self._api_gateway.add_robot_controller(
            cell=self._cell_id, robot_controller=robot_controller, timeout=add_timeout
        )

        await self._api_gateway.wait_for_controller_ready(
            cell=self._cell_id, name=robot_controller.name, timeout=wait_for_ready_timeout
        )

        return self._create_controller(robot_controller.name)

    async def ensure_controller(
        self,
        robot_controller: api.models.RobotController,
        add_timeout: int = DEFAULT_ADD_CONTROLLER_TIMEOUT,
        wait_for_ready_timeout: int = DEFAULT_WAIT_FOR_READY_TIMEOUT,
    ) -> Controller:
        """
        Ensure that a robot controller is added to the cell. If it already exists, it will be returned.
        If it doesn't exist, it will be added and waited for to be ready.
        Args:
            robot_controller (api.models.RobotController): The robot controller to add. You can use helper functions from nova to create these configs easily,
                see :func:`nova.abb_controller`, :func:`nova.fanuc_controller`, :func:`nova.kuka_controller`,
                :func:`nova.universal_robots_controller`, :func:`nova.virtual_controller`, :func:`nova.yaskawa_controller`.
            add_timeout (int): The time to wait for the controller to be added (default: 25).
            wait_for_ready_timeout (int): The time to wait for the controller to be ready (default: 25).

        Returns:
            Controller: The added Controller object.
        """
        controller = await self._api_gateway.get_controller_instance(
            cell=self.cell_id, name=robot_controller.name
        )

        if controller:
            return self._create_controller(controller.controller)
        return await self.add_controller(
            robot_controller, add_timeout=add_timeout, wait_for_ready_timeout=wait_for_ready_timeout
        )

    async def controllers(self) -> list[Controller]:
        """
        List all controllers for this cell.
        Returns:
            list[Controller]: A list of Controller objects associated with this cell.
        """
        instances = await self._api_gateway.list_controllers(cell=self._cell_id)
        return [self._create_controller(ci.controller) for ci in instances]

    async def controller(self, name: str) -> Controller:
        """
        Retrieve a specific controller by name.
        Args:
            name (str): The name of the controller.
        Returns:
            Controller: The Controller object.
        Raises:
            ControllerNotFound: If no controller with the specified name exists.
        """
        controller_instance = await self._api_gateway.get_controller_instance(
            cell=self._cell_id, name=name
        )
        if not controller_instance:
            raise ControllerNotFound(controller=name)
        return self._create_controller(controller_instance.controller)

    async def delete_robot_controller(self, name: str, timeout: int = 25):
        """
        Delete a robot controller from the cell.
        Args:
            name (str): The name of the controller to delete.
            timeout (int): The time to wait for the controller deletion to complete (default: 25).
        """
        await self._api_gateway.delete_robot_controller(
            cell=self._cell_id, controller=name, completion_timeout=timeout
        )

    async def get_robot_cell(self) -> RobotCell:
        """
        Return a RobotCell object containing all known controllers.
        Returns:
            RobotCell: A RobotCell initialized with the available controllers.
        """
        controllers = await self.controllers()
        return RobotCell(timer=None, **{controller.id: controller for controller in controllers})



from nova.actions.base import Action
from nova.actions.container import CombinedActions, MovementController, MovementControllerContext
from nova.actions.io import io_write
from nova.actions.motions import (
    cartesian_ptp,
    cir,
    circular,
    collision_free,
    jnt,
    joint_ptp,
    lin,
    linear,
    ptp,
)

__all__ = [
    "Action",
    "cartesian_ptp",
    "ptp",
    "circular",
    "cir",
    "collision_free",
    "CombinedActions",
    "io_write",
    "joint_ptp",
    "jnt",
    "linear",
    "lin",
    "MovementController",
    "MovementControllerContext",
]



import asyncio

import numpy as np
import rerun as rr
import trimesh
from wandelbots_api_client.models.all_joint_positions_request import AllJointPositionsRequest
from wandelbots_api_client.models.all_joint_positions_response import AllJointPositionsResponse

from nova import MotionSettings
from nova.actions import cartesian_ptp
from nova.api import models
from nova.core.exceptions import PlanTrajectoryFailed
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge


def log_mesh_to_rerun(scene: trimesh.Trimesh) -> None:
    """Log mesh to rerun visualization."""
    vertices = scene.vertices
    faces = scene.faces
    vertex_normals = scene.vertex_normals
    vertex_colors = np.ones((len(vertices), 3), dtype=np.float32)

    rr.log(
        "motion/welding_benchmark",
        rr.Mesh3D(
            vertex_positions=vertices,
            triangle_indices=faces,
            vertex_normals=vertex_normals,
            albedo_factor=vertex_colors,
        ),
        static=True,
    )


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        scene = trimesh.load_mesh(
            "nova_rerun_bridge/example_data/Welding_Benchmark_USA_01.stl", file_type="stl"
        )

        # Define position for the welding part
        mesh_pose = Pose((1100, 0, 0, 0, 0, 0))  # in front of robot, on floor

        # Create transformation matrix from Pose2
        transform = np.eye(4)
        transform[:3, 3] = np.array(mesh_pose.position, dtype=np.float64)
        scene.apply_transform(transform)

        N = 200
        points, face_index = trimesh.sample.sample_surface_even(scene, count=N, radius=1)
        normals = scene.face_normals[face_index]

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            tcp = "Flange"

            # Store points and their configurations count
            tested_points = []
            colors = []

            for point, normal in zip(points, normals, strict=False):
                # Convert normal to rotation vector
                rotation = np.cross([0, 0, 1], normal)  # cross product with Z axis
                if np.any(rotation):  # if not parallel to Z
                    angle = np.arccos(normal[2])  # angle with Z axis
                    rotation = rotation / np.linalg.norm(rotation) * angle
                else:
                    rotation = np.zeros(3)  # no rotation needed

                tested_points.append(point)
                try:
                    response: AllJointPositionsResponse = await nova._api_client.motion_group_kinematic_api.calculate_all_inverse_kinematic(
                        cell=cell.cell_id,
                        motion_group=motion_group.motion_group_id,
                        all_joint_positions_request=AllJointPositionsRequest(
                            motion_group=motion_group.motion_group_id,
                            tcp_pose=models.TcpPose(
                                position=models.Vector3d(x=point[0], y=point[1], z=point[2]),
                                orientation=models.Vector3d(
                                    x=rotation[0], y=rotation[1], z=rotation[2]
                                ),
                                tcp=tcp,
                            ),
                        ),
                    )
                    valid_configs = len(response.joint_positions)
                except Exception:
                    valid_configs = 0

                # Red if unreachable, green gradient based on number of configurations
                if valid_configs == 0:
                    colors.append([1.0, 0.0, 0.0, 1.0])  # Red
                else:
                    intensity = valid_configs / 8.0  # Normalize to [0,1]
                    colors.append([0.0, intensity, 0.0, 1.0])  # Green gradient

            # Log points with colors
            rr.log(
                "motion/reachability", rr.Points3D(tested_points, colors=np.array(colors), radii=3)
            )
            log_mesh_to_rerun(scene)

            home = await motion_group.tcp_pose(tcp)

            actions = [cartesian_ptp(home)]

            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            try:
                joint_trajectory = await motion_group.plan(actions, tcp)
                await bridge.log_actions(actions)
                await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
            except PlanTrajectoryFailed as e:
                await bridge.log_actions(actions)
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)
                return


if __name__ == "__main__":
    asyncio.run(test())



import asyncio
import json

from nova import MotionSettings, api
from nova.actions import cartesian_ptp, joint_ptp
from nova.api import models
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge

TOOL_ASSET = "nova_rerun_bridge/example_data/tool.stl"
tcp_config_dict = {
    "id": "vacuum",
    "readable_name": "vacuum",
    "position": {"x": 0, "y": -80, "z": 340},
    "rotation": {"angles": [0, 0, 0, 0], "type": "ROTATION_VECTOR"},
}


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()
        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # Connect to the controller and activate motion groups
        motion_group_idx = 0
        async with controller[motion_group_idx] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            # Define home
            home_joints = await motion_group.joints()

            # Define new TCP on virtual robot
            tcp_id = tcp_config_dict["id"]
            tcp_config = api.models.RobotTcp.from_json(json.dumps(tcp_config_dict))
            await nova._api_client.virtual_robot_setup_api.add_virtual_robot_tcp(
                cell.cell_id, controller.controller_id, motion_group_idx, tcp_config
            )

            # Wait for tcp configuration to be applied
            while True:
                try:
                    await motion_group.tcp_pose(tcp_id)
                    break
                except api.exceptions.NotFoundException:
                    await asyncio.sleep(0.5)

            # Get current TCP pose and offset it slightly along the x-axis
            current_pose = await motion_group.tcp_pose(tcp_id)
            target_pose = current_pose @ Pose((0, 0, 500, 0, -1.75, 0))

            actions = [joint_ptp(home_joints), cartesian_ptp(target_pose), joint_ptp(home_joints)]

            # you can update the settings of the action
            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            joint_trajectory = await motion_group.plan(actions, tcp_id)

            await bridge.log_actions(actions)
            await bridge.log_trajectory(
                joint_trajectory, tcp_id, motion_group, tool_asset=TOOL_ASSET
            )


if __name__ == "__main__":
    asyncio.run(test())



import asyncio

import numpy as np
import rerun as rr
import trimesh

from nova import MotionSettings
from nova.actions import cartesian_ptp, joint_ptp
from nova.api import models
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.consts import TIME_INTERVAL_NAME


async def test():
    """
    This example demonstrates how to import a PLY file and extract point cloud data.
    We choose the first green point and move the robot to it.
    """
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        # Load PLY file
        mesh = trimesh.load("nova_rerun_bridge/example_data/bin_everything_05.ply")

        # Extract vertex positions and colors
        positions = np.array(mesh.vertices)

        # Point cloud is oriented in a way that it needs to be rotated and translated
        rotation = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        positions = positions @ rotation
        translation = np.array([0, -500, 1200])
        positions = positions + translation

        colors = mesh.visual.vertex_colors[:, :3]  # RGB only, drop alpha

        # Log point cloud
        rr.set_time_seconds(TIME_INTERVAL_NAME, 0)
        rr.log("motion/pointcloud", rr.Points3D(positions, colors=colors))

        # Find green points (high G, low R/B values)
        green_mask = (colors[:, 1] > 100) & (colors[:, 0] < 100) & (colors[:, 2] < 100)
        green_points = positions[green_mask]

        if len(green_points) == 0:
            print("No green points found!")
            return

        # Select first green point
        green_target_point = green_points[0]

        rr.log("motion/target", rr.Points3D([green_target_point], radii=[10], colors=[[0, 255, 0]]))

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            home_joints = await motion_group.joints()
            tcp_names = await motion_group.tcp_names()
            tcp = tcp_names[0]

            greenTargetPose = Pose(
                (green_target_point[0], green_target_point[1], green_target_point[2], np.pi, 0, 0)
            )
            actions = [
                joint_ptp(home_joints),
                cartesian_ptp(greenTargetPose),
                joint_ptp(home_joints),
            ]

            # you can update the settings of the action
            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            joint_trajectory = await motion_group.plan(actions, tcp)

            await bridge.log_actions(actions)
            await bridge.log_trajectory(joint_trajectory, tcp, motion_group)


if __name__ == "__main__":
    asyncio.run(test())



import asyncio
import datetime
import json
import os
import sys
import tempfile
import traceback
from pathlib import Path

from loguru import logger


class SandboxedProgramRunner:
    def __init__(self, program_text: str):
        self.program_text = program_text
        self.project_dir = Path(tempfile.mkdtemp())
        self.program_file = self.project_dir / "program.py"

    async def install_uv(self):
        """Install uv if not already installed."""
        try:
            process = await asyncio.create_subprocess_exec(
                "uv", "--version", stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            await process.communicate()

            if process.returncode == 0:
                logger.info("uv is already installed")
                return

        except FileNotFoundError:
            logger.info("Installing uv...")
            process = await asyncio.create_subprocess_exec(
                sys.executable,
                "-m",
                "pip",
                "install",
                "uv",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await process.communicate()

            if process.returncode != 0:
                raise RuntimeError(f"Failed to install uv: {stderr.decode()}")

            logger.info("uv installed successfully")

    async def validate_program(self):
        """Validate that the program has a main function with correct parameters."""
        import ast

        tree = ast.parse(self.program_text)
        main_func = None
        has_correct_decorator = False

        # Find the main function and check its decorators
        for node in ast.walk(tree):
            if isinstance(node, ast.AsyncFunctionDef) and node.name == "main":
                main_func = node
                # Check decorators
                for decorator in node.decorator_list:
                    if isinstance(decorator, ast.Name) and decorator.id == "program":
                        has_correct_decorator = True
                        break
                    elif isinstance(decorator, ast.Attribute) and decorator.attr == "program":
                        has_correct_decorator = True
                        break

        if not main_func:
            raise ValueError("Program must have an async main function")

        if not has_correct_decorator:
            raise ValueError(
                "Main function must be decorated with @nova.program or @program (imported from nova)"
            )

        logger.info("Program validation successful")

    async def setup_environment(self):
        """Create program file and make it executable."""
        logger.info(f"Setting up program at {self.program_file}")

        # Validate program structure
        await self.validate_program()

        # Ensure uv is installed
        await self.install_uv()

        # Write program file
        self.program_file.write_text(self.program_text)
        self.program_file.chmod(0o755)  # equivalent to chmod u+x

        logger.info("Environment setup complete")

    async def run_program(self, parameters: dict):
        """Run the program using uv with the given parameters."""
        try:
            # Convert parameters to environment variables
            env = os.environ.copy()
            env["PYTHONPATH"] = str(self.project_dir)
            if (api := os.getenv("NOVA_API")) is not None:
                env["NOVA_API"] = api
            if (token := os.getenv("NOVA_ACCESS_TOKEN")) is not None:
                env["NOVA_ACCESS_TOKEN"] = token

            self.program_file.write_text(self.program_text)

            # Convert parameters to command line arguments
            args = []
            for key, value in parameters.items():
                if isinstance(value, (dict, list)):
                    # For complex types, convert to JSON string
                    args.extend([f"--{key}", json.dumps(value)])
                else:
                    # For simple types, convert to string
                    args.extend([f"--{key}={str(value)}"])

            # Build the command for logging
            cmd = ["uv", "run", "--script", str(self.program_file)] + args
            logger.info(f"Running command: {' '.join(cmd)}")

            # Run the program using uv
            process = await asyncio.create_subprocess_exec(
                *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE, env=env
            )

            # Start timing
            start_time = datetime.datetime.now()

            # Create tasks to read stdout and stderr streams
            async def read_stream(stream, prefix):
                while True:
                    line = await stream.readline()
                    if not line:
                        break
                    logger.info(f"{prefix}: {line.decode().strip()}")

            # Run both stream readers concurrently
            await asyncio.gather(
                read_stream(process.stdout, "stdout"), read_stream(process.stderr, "stderr")
            )

            # Wait for process to complete
            return_code = await process.wait()

            end_time = datetime.datetime.now()
            execution_time = end_time - start_time
            logger.info(f"Program execution time: {execution_time}")

            if return_code != 0:
                error_msg = f"Program failed with exit code {return_code}"
                raise RuntimeError(error_msg)

            logger.info("Program execution completed successfully")

        except Exception as e:
            error_msg = f"Error running program: {str(e)}\nTraceback:\n{traceback.format_exc()}"
            logger.error(error_msg)
            raise RuntimeError(error_msg) from e

    async def cleanup(self):
        """Clean up the temporary environment."""
        try:
            import shutil

            shutil.rmtree(self.project_dir)
            logger.info("Cleaned up temporary environment")
        except Exception as e:
            logger.error(f"Error during cleanup: {e}")

    @classmethod
    async def run(cls, program_text: str, parameters: dict):
        """Create a runner, execute the program, and clean up."""
        runner = cls(program_text)
        try:
            await runner.setup_environment()
            await runner.run_program(parameters)
        finally:
            await runner.cleanup()


# Server endpoint
async def run_program_endpoint(program_text: str, parameters: dict):
    """REST endpoint handler for running programs."""
    try:
        logger.info(f"Running program with parameters: {parameters}")
        await SandboxedProgramRunner.run(program_text, parameters)
        return {"status": "success"}
    except Exception as e:
        error_msg = f"Failed to run program: {str(e)}\nTraceback:\n{traceback.format_exc()}"
        logger.error(error_msg)
        return {"status": "error", "message": str(e), "traceback": traceback.format_exc()}



from dataclasses import dataclass

from decouple import config


@dataclass
class Auth0Config:
    """Configuration for Auth0 authentication"""

    domain: str | None = None
    client_id: str | None = None
    audience: str | None = None

    @classmethod
    def from_env(cls) -> "Auth0Config":
        """Create Auth0Config from environment variables"""
        return cls(
            domain=config("NOVA_AUTH0_DOMAIN", default="auth.portal.wandelbots.io"),
            client_id=config("NOVA_AUTH0_CLIENT_ID", default="J7WJUi38xVQdJAEBNRT9Xw1b0fXDb4J2"),
            audience=config("NOVA_AUTH0_AUDIENCE", default="nova-api"),
        )

    def is_complete(self) -> bool:
        """Check if all required fields are set and not None"""
        return bool(self.domain and self.client_id and self.audience)

    def get_validated_config(self) -> tuple[str, str, str]:
        """Get validated config values, ensuring they are not None"""
        if not self.is_complete():
            raise ValueError("Auth0 configuration is incomplete")
        return self.domain, self.client_id, self.audience  # type: ignore



from pathlib import Path
from typing import Any

import yaml


def get_module_path() -> Path:
    return Path(__file__).resolve().parent


def get_dataset_path() -> Path:
    return get_module_path() / "dataset"


def demo_raw() -> dict[str, Any]:
    path = get_dataset_path() / "demo_set.yaml"
    with open(path) as f:
        return yaml.safe_load(f)


def motion_benchmaker_raw() -> dict[str, Any]:
    path = get_dataset_path() / "mb_set.yaml"
    with open(path) as f:
        return yaml.safe_load(f)


def mpinets_raw() -> dict[str, Any]:
    path = get_dataset_path() / "mpinets_set.yaml"
    with open(path) as f:
        return yaml.safe_load(f)



import os
from pathlib import Path


def find_dotenv(filename: str = ".env") -> str:
    """Find .env file in current or parent directories."""
    path = os.getcwd()
    while True:
        env_path = os.path.join(path, filename)
        if os.path.isfile(env_path):
            return env_path
        parent = os.path.dirname(path)
        if parent == path:
            return ".env"
        path = parent


def set_key(key: str, value: str) -> None:
    """Set or update a key in the .env file."""
    env_path = Path(find_dotenv())
    if not env_path.exists():
        env_path.write_text(f'{key}="{value}"\n')
        return

    lines = env_path.read_text().splitlines()
    key_found = False

    for i, line in enumerate(lines):
        if line.startswith(f"{key}="):
            lines[i] = f'{key}="{value}"'
            key_found = True
            break

    if not key_found:
        lines.append(f'{key}="{value}"')

    env_path.write_text("\n".join(lines) + "\n")



import importlib

try:
    from importlib import metadata

    version = metadata.version("wandelbots-nova")
except importlib.metadata.PackageNotFoundError:
    # fallback if not installed in dev
    version = "0.0.0-dev"



import argparse
import inspect
import json
from collections.abc import Callable, Mapping
from typing import (
    Annotated,
    Any,
    Generic,
    ParamSpec,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from docstring_parser import Docstring
from docstring_parser import parse as parse_docstring
from pydantic import BaseModel, Field, PrivateAttr, RootModel, create_model, validate_call
from pydantic.fields import FieldInfo
from pydantic.json_schema import JsonSchemaValue, models_json_schema

Parameters = ParamSpec("Parameters")
Return = TypeVar("Return")


class Function(BaseModel, Generic[Parameters, Return]):
    _wrapped: Callable[Parameters, Return] = PrivateAttr(  # type: ignore
        default_factory=lambda: lambda *args, **kwargs: None
    )
    name: str
    description: str | None
    input: type[BaseModel]
    output: type[BaseModel]

    @classmethod
    def validate(cls, value: Callable[Parameters, Return]) -> "Function[Parameters, Return]":
        if isinstance(value, Function):
            return value
        if not callable(value):
            raise TypeError("value must be callable")

        name = value.__name__

        docstring = parse_docstring(value.__doc__ or "")
        description = docstring.description

        input, output = input_and_output_types(value, docstring)

        function = cls(name=name, description=description, input=input, output=output)
        function._wrapped = validate_call(validate_return=True)(value)
        return function

    def __call__(self, *args: Parameters.args, **kwargs: Parameters.kwargs) -> Return:  # pylint: disable=no-member
        return self._wrapped(*args, **kwargs)

    @property
    def input_schema(self) -> dict[str, Any]:
        return self.input.model_json_schema()

    @property
    def output_schema(self) -> dict[str, Any]:
        return self.output.model_json_schema()

    @property
    def json_schema(self, title: str | None = None) -> JsonSchemaValue:
        _, top_level_schema = models_json_schema(
            [(self.input, "validation"), (self.output, "validation")], title=title or self.name
        )
        return top_level_schema

    def __repr__(self) -> str:
        input_fields = ", ".join(
            f"{k}: {v.annotation.__name__}"  # type: ignore
            for k, v in self.input.model_fields.items()
        )

        # Get the actual output type from RootModel
        if hasattr(self.output, "model_fields") and "root" in self.output.model_fields:
            root_annotation = self.output.model_fields["root"].annotation
            # If it's a TypeVar, get its bound type
            if hasattr(root_annotation, "__bound__") and root_annotation.__bound__:  # type: ignore
                output_type = root_annotation.__bound__.__name__  # type: ignore
            else:
                output_type = root_annotation.__name__  # type: ignore
        else:
            output_type = self.output.__name__

        desc_part = f", description='{self.description}'" if self.description else ""
        return (
            f"Function(name='{self.name}'{desc_part}, input=({input_fields}), output={output_type})"
        )

    def create_parser(self) -> argparse.ArgumentParser:
        """Create an argument parser based on the function's input model.

        Returns:
            argparse.ArgumentParser: A parser configured with arguments matching the input model fields.
        """
        parser = argparse.ArgumentParser(description=self.description or self.name)

        for name, field in self.input.model_fields.items():
            # Convert field type to appropriate Python type
            field_type = field.annotation
            if hasattr(field_type, "__origin__") and field_type.__origin__ is Annotated:  # type: ignore
                field_type = field_type.__origin__  # type: ignore

            # Handle optional fields
            is_optional = False
            if hasattr(field_type, "__origin__") and field_type.__origin__ is Union:  # type: ignore
                field_type = field_type.__args__[0]  # type: ignore
                is_optional = True

            # For complex types (like Pydantic models), use JSON parsing
            if isinstance(field_type, type) and issubclass(field_type, BaseModel):

                def json_type(value: str) -> Any:
                    try:
                        return json.loads(value)
                    except json.JSONDecodeError as e:
                        raise argparse.ArgumentTypeError(f"Invalid JSON for {name}: {e}")

                parser.add_argument(
                    f"--{name}",
                    dest=name,
                    type=json_type,
                    default=field.default if field.default is not None else None,
                    required=not is_optional and field.default is None,
                    help=field.description or f"{name} parameter (JSON format)",
                )
            else:
                # Add argument to parser
                parser.add_argument(
                    f"--{name}",
                    dest=name,
                    type=field_type,  # type: ignore
                    default=field.default if field.default is not None else None,
                    required=not is_optional and field.default is None,
                    help=field.description or f"{name} parameter",
                )

        return parser


def input_and_output_types(
    func: Callable, docstring: Docstring
) -> tuple[type[BaseModel], type[BaseModel]]:
    signature = inspect.signature(func)
    input_types = get_type_hints(func)
    output_type = input_types.pop("return", None)

    input_field_definitions: Mapping[str, Any] = {}
    for order, (name, parameter) in enumerate(signature.parameters.items()):
        default: FieldInfo = (
            Field(...)
            if parameter.default is parameter.empty
            else (
                parameter.default
                if isinstance(parameter.default, FieldInfo)
                else Field(parameter.default)
            )
        )

        # Add field order
        default.json_schema_extra = {"x-order": order}

        # Add description from docstring if available
        if not default.description:
            if param_doc := next((p for p in docstring.params if p.arg_name == name), None):
                default.description = param_doc.description

        input_field_definitions[name] = (parameter.annotation, default)  # type: ignore
    input = create_model("Input", **input_field_definitions, __module__=func.__module__)

    if output_type and isinstance(output_type, type) and issubclass(output_type, BaseModel):
        output = output_type
    else:
        T = TypeVar("T")
        description = None

        # Check if return type is Annotated
        if get_origin(signature.return_annotation) is Annotated:
            base_type, *annotations = get_args(signature.return_annotation)
            for annotation in annotations:
                if isinstance(annotation, FieldInfo):
                    description = annotation.description
                    break
            T = TypeVar("T", bound=base_type)  # type: ignore
        else:
            T = TypeVar("T", bound=output_type)  # type: ignore
            if docstring.returns:
                description = docstring.returns.description

        class Output(RootModel[T]):  # pylint: disable=redefined-outer-name
            root: T = Field(..., description=description)

        output = Output

    return input, output


def wrap(function: Callable[Parameters, Return]) -> Function[Parameters, Return]:
    return Function.validate(function)



from typing import Any, Literal

import pydantic

from nova import api
from nova.actions.base import Action


class WriteAction(Action):
    type: Literal["Write"] = "Write"
    key: str
    value: Any
    device_id: str | None

    @pydantic.model_serializer
    def serialize_model(self):
        return api.models.IOValue(io=self.key, boolean_value=self.value).model_dump()

    def is_motion(self) -> bool:
        return False


def io_write(key: str, value: Any, device_id: str | None = None) -> WriteAction:
    """Create a WriteAction

    Args:
        key: The key to write
        value: The value to write
        device_id: The device id

    Returns:
        The WriteAction

    """
    return WriteAction(key=key, value=value, device_id=device_id)


class ReadAction(Action):
    type: Literal["Read"] = "Read"
    key: str
    device_id: str

    @pydantic.model_serializer
    def serialize_model(self):
        return super().model_dump()

    def is_motion(self) -> bool:
        return False


# TODO: Could move to WS if program representation is not in nova
class CallAction(Action):
    type: Literal["Call"] = "Call"
    device_id: str
    key: str
    arguments: list

    @pydantic.model_serializer
    def serialize_model(self):
        return super().model_dump()

    def is_motion(self) -> bool:
        return False


# TODO: Could move to WS if program representation is not in nova
class ReadPoseAction(Action):
    type: Literal["ReadPose"] = "ReadPose"
    device_id: str
    tcp: str | None = None

    @pydantic.model_serializer
    def serialize_model(self):
        return super().model_dump()

    def is_motion(self) -> bool:
        return False


# TODO: Could move to WS if program representation is not in nova
class ReadJointsAction(Action):
    type: Literal["ReadJoints"] = "ReadJoints"
    device_id: str

    @pydantic.model_serializer
    def serialize_model(self):
        return super().model_dump()

    def is_motion(self) -> bool:
        return False



from typing import Any

import numpy as np
import rerun as rr
import trimesh
from scipy.spatial.transform import Rotation

from nova.api import models
from nova_rerun_bridge import colors
from nova_rerun_bridge.conversion_helpers import normalize_pose
from nova_rerun_bridge.hull_visualizer import HullVisualizer


def log_collision_scenes(collision_scenes: dict[str, models.CollisionScene]):
    for scene_id, scene in collision_scenes.items():
        entity_path = f"collision_scenes/{scene_id}"
        if scene.colliders:
            for collider_id, collider in scene.colliders.items():
                log_colliders_once(entity_path, {collider_id: collider})


def log_colliders_once(entity_path: str, colliders: dict[str, models.Collider]):
    for collider_id, collider in colliders.items():
        pose = normalize_pose(collider.pose)

        if collider.shape.actual_instance is None:
            return

        if isinstance(collider.shape.actual_instance, models.Sphere2):
            # Convert rotation vector to axis-angle format
            if pose.orientation is None:
                continue
            rot_vec = np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            angle = np.linalg.norm(rot_vec)
            if angle > 0:
                axis = rot_vec / angle
            else:
                axis = np.array([0.0, 0.0, 1.0])
                angle: float = 0.0

            rr.log(
                f"{entity_path}/{collider_id}",
                rr.Ellipsoids3D(
                    radii=[
                        collider.shape.actual_instance.radius,
                        collider.shape.actual_instance.radius,
                        collider.shape.actual_instance.radius,
                    ],
                    centers=[[pose.position.x, pose.position.y, pose.position.z]],
                    rotation_axis_angles=[[*axis, angle]],
                    colors=[(221, 193, 193, 255)],
                ),
                static=True,
            )

        elif isinstance(collider.shape.actual_instance, models.RectangularCapsule2):
            # Get parameters from the capsule
            radius = collider.shape.actual_instance.radius
            size_x = collider.shape.actual_instance.sphere_center_distance_x
            size_y = collider.shape.actual_instance.sphere_center_distance_y

            # Create sphere centers at the corners
            sphere_centers = np.array(
                [
                    [size_x / 2, size_y / 2, 0],  # top right
                    [size_x / 2, -size_y / 2, 0],  # bottom right
                    [-size_x / 2, size_y / 2, 0],  # top left
                    [-size_x / 2, -size_y / 2, 0],  # bottom left
                ]
            )

            # Create vertices for spheres at each corner
            vertices = []
            for center in sphere_centers:
                # Create a low-poly sphere at each corner
                sphere = trimesh.creation.icosphere(radius=radius, subdivisions=2)
                sphere_verts = sphere.vertices + center
                vertices.extend(sphere_verts)

            # Convert to numpy array for transformation
            vertices = np.array(vertices)

            # Transform vertices to world position
            transform = np.eye(4)
            transform[:3, 3] = [pose.position.x, pose.position.y, pose.position.z]
            rot_mat = Rotation.from_rotvec(
                np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            ).as_matrix()
            transform[:3, :3] = rot_mat
            vertices = np.array([transform @ np.append(v, 1) for v in vertices])[:, :3]

            # Create hull visualization
            polygons = HullVisualizer.compute_hull_outlines_from_points(vertices)

            if polygons:
                line_segments = [p.tolist() for p in polygons]
                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.LineStrips3D(
                        line_segments,
                        radii=rr.Radius.ui_points(0.75),
                        colors=[[221, 193, 193, 255]],
                    ),
                    static=True,
                )

        elif isinstance(collider.shape.actual_instance, models.Rectangle2):
            # Create vertices for a rectangle in XY plane
            half_x = collider.shape.actual_instance.size_x / 2
            half_y = collider.shape.actual_instance.size_y / 2
            vertices = np.array(
                [
                    [-half_x, -half_y, 0],  # bottom left
                    [half_x, -half_y, 0],  # bottom right
                    [half_x, half_y, 0],  # top right
                    [-half_x, half_y, 0],  # top left
                ]
            )

            # Transform vertices
            transform = np.eye(4)
            transform[:3, 3] = [pose.position.x, pose.position.y, pose.position.z]
            rot_mat = Rotation.from_rotvec(
                np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            ).as_matrix()
            transform[:3, :3] = rot_mat
            vertices = np.array([transform @ np.append(v, 1) for v in vertices])[:, :3]

            # Create line segments for the rectangle outline
            line_segments = [
                [vertices[0].tolist(), vertices[1].tolist()],
                [vertices[1].tolist(), vertices[2].tolist()],
                [vertices[2].tolist(), vertices[3].tolist()],
                [vertices[3].tolist(), vertices[0].tolist()],
            ]

            rr.log(
                f"{entity_path}/{collider_id}",
                rr.LineStrips3D(
                    line_segments, radii=rr.Radius.ui_points(0.75), colors=[[221, 193, 193, 255]]
                ),
                static=True,
            )

        elif isinstance(collider.shape.actual_instance, models.Box2):
            # Create rotation matrix from orientation
            rot_mat = Rotation.from_rotvec(
                np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            ).as_matrix()

            # Create box vertices
            half_sizes = [
                collider.shape.actual_instance.size_x / 2,
                collider.shape.actual_instance.size_y / 2,
                collider.shape.actual_instance.size_z / 2,
            ]
            box = trimesh.creation.box(extents=[s * 2 for s in half_sizes])
            vertices = np.array(box.vertices)

            # Transform vertices
            transform = np.eye(4)
            transform[:3, 3] = [pose.position.x, pose.position.y, pose.position.z]
            transform[:3, :3] = rot_mat
            vertices = np.array([transform @ np.append(v, 1) for v in vertices])[:, :3]

            # Create hull visualization
            polygons = HullVisualizer.compute_hull_outlines_from_points(vertices)

            if polygons:
                line_segments = [p.tolist() for p in polygons]
                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.LineStrips3D(
                        line_segments,
                        radii=rr.Radius.ui_points(0.75),
                        colors=[[221, 193, 193, 255]],
                    ),
                    static=True,
                )

        elif isinstance(collider.shape.actual_instance, models.Cylinder2):
            height = collider.shape.actual_instance.height
            radius = collider.shape.actual_instance.radius

            # Generate trimesh capsule
            cylinder = trimesh.creation.cylinder(height=height, radius=radius)

            # Extract vertices and faces for solid visualization
            vertices = np.array(cylinder.vertices)

            # Transform vertices to world position
            transform = np.eye(4)
            transform[:3, 3] = [pose.position.x, pose.position.y, pose.position.z]
            rot_mat = Rotation.from_rotvec(
                np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            )
            transform[:3, :3] = rot_mat.as_matrix()

            vertices = np.array([transform @ np.append(v, 1) for v in vertices])[:, :3]

            polygons = HullVisualizer.compute_hull_outlines_from_points(vertices)

            if polygons:
                line_segments = [p.tolist() for p in polygons]
                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.LineStrips3D(
                        line_segments,
                        radii=rr.Radius.ui_points(0.75),
                        colors=[[221, 193, 193, 255]],
                    ),
                    static=True,
                )

        elif isinstance(collider.shape.actual_instance, models.Capsule2):
            height = collider.shape.actual_instance.cylinder_height
            radius = collider.shape.actual_instance.radius

            # Generate trimesh capsule
            capsule = trimesh.creation.capsule(height=height, radius=radius, count=[6, 8])

            # Extract vertices and faces for solid visualization
            vertices = np.array(capsule.vertices)

            # Transform vertices to world position
            transform = np.eye(4)
            transform[:3, 3] = [pose.position.x, pose.position.y, pose.position.z]
            rot_mat = Rotation.from_rotvec(
                np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            )
            transform[:3, :3] = rot_mat.as_matrix()

            vertices = np.array([transform @ np.append(v, 1) for v in vertices])[:, :3]

            polygons = HullVisualizer.compute_hull_outlines_from_points(vertices)

            if polygons:
                line_segments = [p.tolist() for p in polygons]
                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.LineStrips3D(
                        line_segments,
                        radii=rr.Radius.ui_points(0.75),
                        colors=[[221, 193, 193, 255]],
                    ),
                    static=True,
                )

        elif isinstance(collider.shape.actual_instance, models.ConvexHull2):
            # Transform vertices to world position
            vertices = np.array(collider.shape.actual_instance.vertices)
            transform = np.eye(4)
            transform[:3, 3] = [pose.position.x, pose.position.y, pose.position.z]
            rot_mat = Rotation.from_rotvec(
                np.array([pose.orientation.x, pose.orientation.y, pose.orientation.z])
            ).as_matrix()
            transform[:3, :3] = rot_mat

            # Apply transformation
            vertices = np.array([transform @ np.append(v, 1) for v in vertices])[:, :3]

            polygons = HullVisualizer.compute_hull_outlines_from_points(vertices)

            if polygons:
                line_segments = [p.tolist() for p in polygons]
                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.LineStrips3D(
                        line_segments, radii=rr.Radius.ui_points(1.5), colors=[colors.colors[2]]
                    ),
                    static=True,
                )

                vertices, triangles, normals = HullVisualizer.compute_hull_mesh(polygons)

                rr.log(
                    f"{entity_path}/{collider_id}",
                    rr.Mesh3D(
                        vertex_positions=vertices,
                        triangle_indices=triangles,
                        vertex_normals=normals,
                        albedo_factor=[colors.colors[0]],
                    ),
                    static=True,
                )


def extract_link_chain_and_tcp(
    collision_scenes: dict, motion_group_type: str
) -> tuple[list[Any], list[Any]]:
    """Extract link chain and TCP from collision scenes."""
    # Get first scene (name can vary)
    scene = next(iter(collision_scenes.values()), None)
    if not scene:
        return [], []

    # Try to get motion groups
    motion_group = scene.motion_groups.get(motion_group_type)
    if not motion_group:
        return [], []

    return (getattr(motion_group, "link_chain", []), getattr(motion_group, "tool", []))



from abc import ABC, abstractmethod
from typing import Literal

import pydantic
import wandelbots_api_client as wb

from nova import api
from nova.actions.base import Action
from nova.types.motion_settings import MotionSettings
from nova.types.pose import Pose

PoseOrVectorTuple = (
    Pose | tuple[float, float, float, float, float, float] | tuple[float, float, float]
)


class CollisionFreeMotion(Action):
    """A motion that is collision free"""

    type: Literal["collision_free_ptp"] = "collision_free_ptp"
    target: Pose | tuple[float, ...]
    settings: MotionSettings = MotionSettings()
    collision_scene: wb.models.CollisionScene | None = None

    def _to_api_model(self) -> api.models.PlanCollisionFreePTPRequestTarget:
        return wb.models.PlanCollisionFreePTPRequestTarget(
            self.target._to_wb_pose2() if isinstance(self.target, Pose) else list(self.target)
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()

    def is_motion(self) -> bool:
        return True


def collision_free(
    target: Pose | tuple[float, ...],
    settings: MotionSettings = MotionSettings(),
    collision_scene: wb.models.CollisionScene | None = None,
) -> CollisionFreeMotion:
    return CollisionFreeMotion(target=target, settings=settings, collision_scene=collision_scene)


class Motion(Action, ABC):
    """Base model of a motion

    Args:
        type: the type of the motion
        settings: the settings of the motion

    """

    type: Literal["linear", "cartesian_ptp", "circular", "joint_ptp", "spline"]
    target: Pose | tuple[float, ...]
    settings: MotionSettings = MotionSettings()
    collision_scene: wb.models.CollisionScene | None = None

    @property
    def is_cartesian(self):
        return isinstance(self.target, Pose)

    def is_motion(self) -> bool:
        return True


class UnresolvedMotion(Motion, ABC):
    @abstractmethod
    async def resolve(
        self,
        initial_joints: tuple[float, ...],
        collision_scene: wb.models.CollisionScene | None,
        configuration: dict,
        moving_robot_identifier: str,
    ) -> tuple[list[Motion], tuple[float, ...]] | None:
        """Convert the motion to a list of motion primitives

        Args:
            initial_joints: Joint positions at start of motion
            collision_scene: The collision scene used to check collisions
            configuration: E.g. data of physical setup of robot system, cell, etc.
            moving_robot_identifier: The identifier of the robot that is moving in the scene

        Returns:
            Tuple of resolved motions and the joint position at the end of the motions. None, if the motion can't be resolved

        """


class Linear(Motion):
    """A linear motion

    Examples:
    >>> Linear(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=10))
    Linear(type='linear', target=Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6)), settings=MotionSettings(min_blending_velocity=None, position_zone_radius=None, joint_velocity_limits=None, joint_acceleration_limits=None, tcp_velocity_limit=10.0, tcp_acceleration_limit=None, tcp_orientation_velocity_limit=None, tcp_orientation_acceleration_limit=None), collision_scene=None)

    """

    type: Literal["linear"] = "linear"
    target: Pose

    def _to_api_model(self) -> api.models.PathLine:
        """Serialize the model to the API model

        Examples:
        >>> Linear(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=10))._to_api_model()
        PathLine(target_pose=Pose2(position=[1, 2, 3], orientation=[4, 5, 6]), path_definition_name='PathLine')
        """
        return api.models.PathLine(
            target_pose=api.models.Pose2(**self.target.model_dump()),
            path_definition_name="PathLine",
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def linear(
    target: PoseOrVectorTuple,
    settings: MotionSettings = MotionSettings(),
    collision_scene: wb.models.CollisionScene | None = None,
) -> Linear:
    """Convenience function to create a linear motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        settings: the motion settings

    Returns: the linear motion

    Examples:
    >>> ms = MotionSettings(tcp_velocity_limit=10)
    >>> assert linear((1, 2, 3, 4, 5, 6), settings=ms) == Linear(target=Pose((1, 2, 3, 4, 5, 6)), settings=ms)
    >>> assert linear((1, 2, 3)) == linear((1, 2, 3, 0, 0, 0))
    >>> assert linear(Pose((1, 2, 3, 4, 5, 6)), settings=ms) == linear((1, 2, 3, 4, 5, 6), settings=ms)

    """
    if not isinstance(target, Pose):
        t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
        target = Pose(t)

    return Linear(target=target, settings=settings, collision_scene=collision_scene)


lin = linear


class CartesianPTP(Motion):
    """A point-to-point motion

    Examples:
    >>> CartesianPTP(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=30))
    CartesianPTP(type='cartesian_ptp', target=Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6)), settings=MotionSettings(min_blending_velocity=None, position_zone_radius=None, joint_velocity_limits=None, joint_acceleration_limits=None, tcp_velocity_limit=30.0, tcp_acceleration_limit=None, tcp_orientation_velocity_limit=None, tcp_orientation_acceleration_limit=None), collision_scene=None)

    """

    type: Literal["cartesian_ptp"] = "cartesian_ptp"

    def _to_api_model(self) -> api.models.PathCartesianPTP:
        """Serialize the model to the API model

        Examples:
        >>> CartesianPTP(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=30))._to_api_model()
        PathCartesianPTP(target_pose=Pose2(position=[1, 2, 3], orientation=[4, 5, 6]), path_definition_name='PathCartesianPTP')
        """
        if not isinstance(self.target, Pose):
            raise ValueError("Target must be a Pose object")
        return api.models.PathCartesianPTP(
            target_pose=api.models.Pose2(**self.target.model_dump()),
            path_definition_name="PathCartesianPTP",
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def cartesian_ptp(
    target: PoseOrVectorTuple,
    settings: MotionSettings = MotionSettings(),
    collision_scene: wb.models.CollisionScene | None = None,
) -> CartesianPTP:
    """Convenience function to create a point-to-point motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        settings: the motion settings

    Returns: the point-to-point motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert cartesian_ptp((1, 2, 3, 4, 5, 6), settings=ms) == CartesianPTP(target=Pose((1, 2, 3, 4, 5, 6)), settings=ms)
    >>> assert cartesian_ptp((1, 2, 3)) == cartesian_ptp((1, 2, 3, 0, 0, 0))
    >>> assert cartesian_ptp(Pose((1, 2, 3, 4, 5, 6)), settings=ms) == cartesian_ptp((1, 2, 3, 4, 5, 6), settings=ms)

    """
    if not isinstance(target, Pose):
        t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
        target = Pose(t)

    return CartesianPTP(target=target, settings=settings, collision_scene=collision_scene)


ptp = cartesian_ptp


class Circular(Motion):
    """A circular motion

    Args:
        intermediate: the intermediate pose

    """

    type: Literal["circular"] = "circular"
    intermediate: Pose

    def _to_api_model(self) -> api.models.PathCircle:
        """Serialize the model to a dictionary

        Examples:
        >>> Circular(target=Pose((1, 2, 3, 4, 5, 6)), intermediate=Pose((10, 20, 30, 40, 50, 60)), settings=MotionSettings(tcp_velocity_limit=30))._to_api_model()
        PathCircle(via_pose=Pose2(position=[10, 20, 30], orientation=[40, 50, 60]), target_pose=Pose2(position=[1, 2, 3], orientation=[4, 5, 6]), path_definition_name='PathCircle')
        """
        if not isinstance(self.target, Pose):
            raise ValueError("Target must be a Pose object")
        if not isinstance(self.intermediate, Pose):
            raise ValueError("Intermediate must be a Pose object")
        return api.models.PathCircle(
            target_pose=api.models.Pose2(**self.target.model_dump()),
            via_pose=api.models.Pose2(**self.intermediate.model_dump()),
            path_definition_name="PathCircle",
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def circular(
    target: PoseOrVectorTuple,
    intermediate: PoseOrVectorTuple,
    settings: MotionSettings = MotionSettings(),
    collision_scene: wb.models.CollisionScene | None = None,
) -> Circular:
    """Convenience function to create a circular motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        intermediate: the intermediate pose or vector. If the intermediate is a vector, the orientation is set to
            (0, 0, 0).
        settings: the motion settings

    Returns: the circular motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert circular((1, 2, 3, 4, 5, 6), (7, 8, 9, 10, 11, 12), settings=ms) == Circular(target=Pose((1, 2, 3, 4, 5, 6)), intermediate=Pose((7, 8, 9, 10, 11, 12)), settings=ms)
    >>> assert circular((1, 2, 3), (4, 5, 6)) == circular((1, 2, 3, 0, 0, 0), (4, 5, 6, 0, 0, 0))

    """
    if not isinstance(target, Pose):
        t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
        target = Pose(t)

    if not isinstance(intermediate, Pose):
        i = (*intermediate, 0.0, 0.0, 0.0) if len(intermediate) == 3 else intermediate
        intermediate = Pose(i)

    return Circular(
        target=target, intermediate=intermediate, settings=settings, collision_scene=collision_scene
    )


cir = circular


class JointPTP(Motion):
    """A joint PTP motion

    Examples:
    >>> JointPTP(target=(1, 2, 3, 4, 5, 6), settings=MotionSettings(tcp_velocity_limit=30))
    JointPTP(type='joint_ptp', target=(1.0, 2.0, 3.0, 4.0, 5.0, 6.0), settings=MotionSettings(min_blending_velocity=None, position_zone_radius=None, joint_velocity_limits=None, joint_acceleration_limits=None, tcp_velocity_limit=30.0, tcp_acceleration_limit=None, tcp_orientation_velocity_limit=None, tcp_orientation_acceleration_limit=None), collision_scene=None)

    """

    type: Literal["joint_ptp"] = "joint_ptp"

    def _to_api_model(self) -> api.models.PathJointPTP:
        """Serialize the model to the API model

        Examples:
        >>> JointPTP(target=(1, 2, 3, 4, 5, 6, 7), settings=MotionSettings(tcp_velocity_limit=30))._to_api_model()
        PathJointPTP(target_joint_position=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0], path_definition_name='PathJointPTP')
        """
        if not isinstance(self.target, tuple):
            raise ValueError("Target must be a tuple object")
        return api.models.PathJointPTP(
            target_joint_position=list(self.target), path_definition_name="PathJointPTP"
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def joint_ptp(
    target: tuple[float, ...],
    settings: MotionSettings = MotionSettings(),
    collision_scene: wb.models.CollisionScene | None = None,
) -> JointPTP:
    """Convenience function to create a joint PTP motion

    Args:
        target: the target joint configuration
        settings: the motion settings

    Returns: the joint PTP motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert joint_ptp((1, 2, 3, 4, 5, 6), settings=ms) == JointPTP(target=(1, 2, 3, 4, 5, 6), settings=ms)

    """
    return JointPTP(target=target, settings=settings, collision_scene=collision_scene)


jnt = joint_ptp


class Spline(Motion):
    """A spline motion

    Args:
        path_parameter: the path parameter between 0 and 1
        time: the time in seconds

    """

    type: Literal["spline"] = "spline"
    path_parameter: float = pydantic.Field(1, ge=0)
    time: float | None = pydantic.Field(default=None, ge=0)

    @pydantic.model_serializer
    def serialize_model(self):
        """Serialize the model to a dictionary

        Examples:
        >>> JointPTP(target=(1, 2, 3, 4, 5, 6, 7), settings=MotionSettings(tcp_velocity_limit=30)).model_dump()
        {'target_joint_position': [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0], 'path_definition_name': 'PathJointPTP'}
        """
        raise NotImplementedError("Spline motion is not yet implemented")


def spline(
    target: PoseOrVectorTuple,
    settings: MotionSettings = MotionSettings(),
    path_parameter: float = 1,
    time=None,
    collision_scene: wb.models.CollisionScene | None = None,
) -> Spline:
    """Convenience function to create a spline motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        settings: the motion settings
        path_parameter: the path parameter between 0 and 1
        time: the time in seconds

    Returns: the spline motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert spl((1, 2, 3, 4, 5, 6), settings=ms) == Spline(target=Pose((1, 2, 3, 4, 5, 6)), settings=ms)
    >>> assert spl((1, 2, 3)) == spl((1, 2, 3, 0, 0, 0))

    """
    if not isinstance(target, Pose):
        t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
        target = Pose(t)

    return Spline(
        target=target,
        settings=settings,
        path_parameter=path_parameter,
        time=time,
        collision_scene=collision_scene,
    )


spl = spline



from typing import Any

from .client import ClientConfig, OPCUAClient, SubscriptionConfig


async def opcua_write(url: str, node_id: str, value: Any, options: dict | None = None):
    """Write a value to the opcua node

    Node ids should be based on opcua standard.
    More information about how opcua node id string notation works can be found here:
    https://documentation.unified-automation.com/uasdkhp/1.4.1/html/_l2_ua_node_ids.html

    Args:
        url: the url of the opcua server
        node_id: id of the node to write the value
        value: the value to write
        options: configuration for the opcua client
            {
                # the timeout for each communication happening between opcua client and the server
                request_timeout_seconds: int,

                security_config: {
                    # the storage service path for the client certificate
                    client_certificate_path: str,

                    # the storage service path for the client private key
                    client_private_key_path: str,

                    # the security policy to use, e.g. Basic256Sha256
                    security_policy: str,

                    # the message security mode to use, e.g. SignAndEncrypt
                    message_security_mode: str
                }
            }
    """
    options = options or {}
    client_config = ClientConfig(**options)

    async with OPCUAClient(url, client_config) as opc:
        await opc.write_node(node_id, value)


async def opcua_read(url: str, node_id: str, options: dict | None = None) -> Any:
    """Reads the value of a opcua node and returns the result

    Node ids should be based on opcua standard.
    More information about how opcua node id string notation works can be found here:
    https://documentation.unified-automation.com/uasdkhp/1.4.1/html/_l2_ua_node_ids.html

    Args:
        url: the url of the opcua server
        node_id: id of the node to read the value of
        options: configuration for the opcua client
            {
                # the timeout for each communication happening between opcua client and the server
                request_timeout_seconds: int,
                security_config: {
                    # the storage service path for the client certificate
                    client_certificate_path: str,

                    # the storage service path for the client private key
                    client_private_key_path: str,

                    # the security policy to use, e.g. Basic256Sha256
                    security_policy: str,

                    # the message security mode to use, e.g. SignAndEncrypt
                    message_security_mode: str
                }
            }


    Returns:
        the value of the node

    """
    options = options or {}
    client_config = ClientConfig(**options)

    async with OPCUAClient(url, client_config) as opc:
        return await opc.read_node(node_id)


async def opcua_call(url: str, object_id: str, function_id: str, *args) -> Any:
    """executes the opcua function and returns the result

    Node ids should be based on opcua standard.
    More information about how opcua node id string notation works can be found here:
    https://documentation.unified-automation.com/uasdkhp/1.4.1/html/_l2_ua_node_ids.html

    Args:
        url: url of the opcua server
        object_id: node id of the object the function belongs to
        function_id : node id of the function
        *args: the arguments to the function,
               these parameters will be provided to OPCUA server while calling the function you specified
               the last parameter can be a wandelscript record to configure the opcua client.
                {
                    # the timeout for each communication happening between opcua client and the server
                     request_timeout_seconds: int,
                     security_config: {
                        # the storage service path for the client certificate
                        client_certificate_path: str,

                        # the storage service path for the client private key
                        client_private_key_path: str,

                        # the security policy to use, e.g. Basic256Sha256
                        security_policy: str,

                        # the message security mode to use, e.g. SignAndEncrypt
                        message_security_mode: str
                    }
                }


    Returns:
        the value returned by the opcua function
    """
    # there is no args
    if not args:
        opcua_function_args, options = (), {}
    # the last arg is a record
    elif isinstance(args[-1], dict):
        opcua_function_args, options = args[:-1], args[-1]
    # there are args but the last one is not a record
    else:
        opcua_function_args, options = args, {}

    client_config = ClientConfig(**options)
    async with OPCUAClient(url, client_config) as opc:
        return await opc.call_node(object_id, function_id, *opcua_function_args)


async def wait_for_opcua_value(url: str, node_id: str, value: Any, config: dict | None = None):
    """watches the opcua node with the given key until it matches the given value

    Node ids should be based on opcua standard.
    More information about how opcua node id string notation works can be found here:
    https://documentation.unified-automation.com/uasdkhp/1.4.1/html/_l2_ua_node_ids.html

    Args:
        url: url of the opcua server
        node_id: id of the node to watch the value of
        value: value that the node should have
        config: configuration for the subscription
    """
    config = config or {}
    subscription_config = SubscriptionConfig(**config)
    options = ClientConfig(
        request_timeout_seconds=subscription_config.request_timeout_seconds,
        security_config=subscription_config.security_config,
    )

    async with OPCUAClient(url, options) as opc:

        def condition(node_value: Any):
            return node_value == value

        await opc.watch_node_until_condition(node_id, condition, config=subscription_config)



import asyncio

import numpy as np
import rerun as rr
from wandelbots_api_client.models import (
    CoordinateSystem,
    RotationAngles,
    RotationAngleTypes,
    Vector3d,
)

from nova import MotionSettings
from nova.actions.motions import cartesian_ptp, collision_free
from nova.api import models
from nova.core.exceptions import PlanTrajectoryFailed
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge

"""
Example of planning a collision free PTP motion. A sphere is placed in the robot's path and the robot uses collision free p2p to move around it.
"""


async def build_collision_world(
    nova: Nova, cell_name: str, robot_setup: models.OptimizerSetup
) -> str:
    collision_api = nova._api_client.store_collision_components_api
    scene_api = nova._api_client.store_collision_scenes_api

    # define annoying obstacle
    sphere_collider = models.Collider(
        shape=models.ColliderShape(models.Sphere2(radius=100, shape_type="sphere")),
        pose=models.Pose2(position=[-100, -500, 200]),
    )
    await collision_api.store_collider(
        cell=cell_name, collider="annoying_obstacle", collider2=sphere_collider
    )

    # define TCP collider geometry
    tool_collider = models.Collider(
        shape=models.ColliderShape(
            models.Box2(size_x=100, size_y=100, size_z=100, shape_type="box", box_type="FULL")
        )
    )
    await collision_api.store_collision_tool(
        cell=cell_name, tool="tool_box", request_body={"tool_collider": tool_collider}
    )

    # define robot link geometries
    robot_link_colliders = await collision_api.get_default_link_chain(
        cell=cell_name, motion_group_model=robot_setup.motion_group_type
    )
    await collision_api.store_collision_link_chain(
        cell=cell_name, link_chain="robot_links", collider=robot_link_colliders
    )

    # assemble scene
    scene = models.CollisionScene(
        colliders={"annoying_obstacle": sphere_collider},
        motion_groups={
            robot_setup.motion_group_type: models.CollisionMotionGroup(
                tool={"tool_geometry": tool_collider}, link_chain=robot_link_colliders
            )
        },
    )
    scene_id = "collision_scene"
    await scene_api.store_collision_scene(
        cell_name, scene_id, models.CollisionSceneAssembly(scene=scene)
    )
    return scene_id


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur5",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR5E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        await nova._api_client.virtual_robot_setup_api.set_virtual_robot_mounting(
            cell="cell",
            controller=controller.controller_id,
            id=0,
            coordinate_system=CoordinateSystem(
                coordinate_system="world",
                name="mounting",
                reference_uid="",
                position=Vector3d(x=0, y=0, z=0),
                rotation=RotationAngles(
                    angles=[0, 0, 0], type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ
                ),
            ),
        )

        # NC-1047
        await asyncio.sleep(5)

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            tcp = "Flange"

            robot_setup: models.OptimizerSetup = await motion_group._get_optimizer_setup(tcp=tcp)
            robot_setup.safety_setup.global_limits.tcp_velocity_limit = 200

            collision_scene_id = await build_collision_world(nova, "cell", robot_setup)

            await bridge.log_collision_scenes()

            # Use default planner to move to the right of the sphere
            home = await motion_group.tcp_pose(tcp)
            actions = [
                cartesian_ptp(home),
                cartesian_ptp(target=Pose((300, -400, 200, np.pi, 0, 0))),
            ]

            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            try:
                joint_trajectory = await motion_group.plan(
                    actions, tcp, start_joint_position=(0, -np.pi / 2, np.pi / 2, 0, 0, 0)
                )
                await bridge.log_actions(actions)
                await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
            except PlanTrajectoryFailed as e:
                await bridge.log_actions(actions)
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)

            rr.log(
                "motion/target_", rr.Points3D([[-500, -400, 200]], radii=[10], colors=[(0, 255, 0)])
            )

            # Use default planner to move to the left of the sphere
            # -> this will collide
            # only plan don't move
            actions = [cartesian_ptp(target=Pose((-500, -400, 200, np.pi, 0, 0)))]

            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            try:
                joint_trajectory_with_collision = await motion_group.plan(
                    actions, tcp, start_joint_position=joint_trajectory.joint_positions[-1].joints
                )
                await bridge.log_actions(actions)
                await bridge.log_trajectory(joint_trajectory_with_collision, tcp, motion_group)
            except PlanTrajectoryFailed as e:
                await bridge.log_actions(actions)
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)

            # Plan collision free PTP motion around the sphere
            scene_api = nova._api_client.store_collision_scenes_api
            collision_scene = await scene_api.get_stored_collision_scene(
                cell="cell", scene=collision_scene_id
            )

            welding_actions = [
                collision_free(
                    target=Pose((-500, -400, 200, np.pi, 0, 0)),
                    collision_scene=collision_scene,
                    settings=MotionSettings(tcp_velocity_limit=30),
                )
            ]

            trajectory_plan_combined = await motion_group.plan(
                welding_actions,
                tcp=tcp,
                start_joint_position=joint_trajectory.joint_positions[-1].joints,
            )
            await bridge.log_actions(welding_actions)
            await bridge.log_trajectory(trajectory_plan_combined, tcp, motion_group)


if __name__ == "__main__":
    asyncio.run(test())



import asyncio
import os.path
import tempfile
from collections.abc import Callable
from datetime import datetime
from typing import Any

import asyncua
import httpx
import pydantic
from asyncua import ua
from asyncua.common import Node
from asyncua.common.subscription import DataChangeNotif, DataChangeNotificationHandlerAsync
from asyncua.crypto import security_policies
from asyncua.ua import CreateSubscriptionParameters, DataValue, Variant, VariantType
from decouple import config

from nova import logger


class DataChangeSubscription(DataChangeNotificationHandlerAsync):
    """
    This class extends the DataChangeNotificationHandlerAsync class from the asyncua library.
    It is used to create a subscription that listens for data change notifications on a node.
    It supports a condition function that is used to determine when the subscription should be completed.
    """

    def __init__(self, condition: Callable[[Any], bool], print_received_messages=False):
        super().__init__()
        self._flag = asyncio.Event()
        self._condition = condition
        self._print_received_messages = print_received_messages

    async def datachange_notification(self, node: Node, val, data: DataChangeNotif) -> None:
        if self._print_received_messages:
            current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
            print(f"{current_time}: Received data change notification for node {node}, {val}")

        if self._condition(val):
            self._flag.set()

    def __await__(self):
        return self._flag.wait().__await__()


class SecurityConfig(pydantic.BaseModel):
    security_policy: str
    message_security_mode: str
    client_certificate_path: str
    client_private_key_path: str


class ClientConfig(pydantic.BaseModel):
    # this is the default value of the underlying library
    request_timeout_seconds: int = 4
    security_config: SecurityConfig | None = None


class SubscriptionConfig(pydantic.BaseModel):
    requested_publishing_interval: float = 100
    requested_lifetime_count: int = 10000
    max_notifications_per_publish: int = 1000
    priority: int = 0
    queue_size: int = 1
    sampling_interval: float = 0.0
    print_received_messages: bool = False
    request_timeout_seconds: int = 4
    security_config: SecurityConfig | None = None

    @classmethod
    def from_dict(cls, config_dict: dict):
        return cls(**config_dict)


async def fetch_certificate(certificate_path: str, private_key_path: str) -> tuple[bytes, bytes]:
    cell = config("K8S_NAMESPACE", default="cell")

    cert_url = f"http://api-gateway:8080/api/v1/cells/{cell}/store/objects/{certificate_path}"
    private_key_url = (
        f"http://api-gateway:8080/api/v1/cells/{cell}/store/objects/{private_key_path}"
    )

    async with httpx.AsyncClient() as client:
        response = await client.get(cert_url)
        response.raise_for_status()
        cert = response.content

        response = await client.get(private_key_url)
        response.raise_for_status()
        private_key = response.content

    return private_key, cert


async def certificate_files(cert_path: str, private_key_path: str) -> tuple[str, str]:
    private_key, cert = await fetch_certificate(cert_path, private_key_path)
    cert_suffix = os.path.splitext(cert_path)[1]
    private_key_suffix = os.path.splitext(private_key_path)[1]

    with tempfile.NamedTemporaryFile(delete=False, mode="wb", suffix=cert_suffix) as temp_file:
        temp_file.write(cert)
        cert_path = temp_file.name
    with tempfile.NamedTemporaryFile(
        delete=False, mode="wb", suffix=private_key_suffix
    ) as temp_file:
        temp_file.write(private_key)
        private_key_path = temp_file.name

    return private_key_path, cert_path


class OPCUAClient:
    """A wrapper around asyncua client"""

    def __init__(self, url: str, options: ClientConfig = ClientConfig()):
        self._options = options
        self._client = asyncua.Client(url=url, timeout=options.request_timeout_seconds)

    async def _extend_client_with_certificate(self):
        if not self._options.security_config:
            raise Exception("Security configuration is missing")

        # Taken from self._client.set_security_string
        policy_class = getattr(
            security_policies, f"SecurityPolicy{self._options.security_config.security_policy}"
        )
        mode = getattr(ua.MessageSecurityMode, self._options.security_config.message_security_mode)

        self._temp_private_key_file, self._temp_cert_file = await certificate_files(
            self._options.security_config.client_certificate_path,
            self._options.security_config.client_private_key_path,
        )
        await self._client.set_security(
            policy_class, self._temp_cert_file, self._temp_private_key_file, None, None, mode
        )

    def _delete_certificate_files(self):
        files = []
        if hasattr(self, "_temp_cert_file") and os.path.exists(self._temp_cert_file):
            files.append(self._temp_cert_file)
        if hasattr(self, "_temp_private_key_file") and os.path.exists(self._temp_private_key_file):
            files.append(self._temp_private_key_file)

        for file_path in files:
            try:
                os.remove(file_path)
            except Exception as cleanup_err:
                logger.error(f"Error cleaning up temporary file {file_path}: {cleanup_err}")

    async def __aenter__(self):
        try:
            if self._options.security_config:
                await self._extend_client_with_certificate()

            await self._client.connect()
        except:
            self._delete_certificate_files()
            raise
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._delete_certificate_files()
        await self._client.disconnect()

    async def read_node(self, key: str) -> Any:
        """
        Reads the value of a opcua node and returns the result
        Args:
            key: opc ua node id

        Returns: the value of the node

        """
        node = await self.get_node(key)
        val = await node.read_value()
        return val

    async def write_node(self, key: str, val: Any):
        """
        Writes the value to a opcua node
        Args:
            key: opc ua node id
            val: the value to write
        """
        node = await self.get_node(key)
        data_type_variant = await node.read_data_type_as_variant_type()
        data_value = DataValue(Variant(val, data_type_variant))
        await node.write_value(data_value)

    async def call_node(self, parent_key: str, function_key: str, *args):
        """
        Calls a method on a opcua node
        Args:
            parent_key: the parent node id which contains the method
            function_key: the method node id
            *args: arguments to the opc ua method

        Returns:
            the response returned by the opcua method

        """
        node = await self.get_node(parent_key)
        method_node = await self.get_node(function_key)

        mapped_args = await self._map_arguments(method_node, *args)
        result = await node.call_method(method_node.nodeid, *mapped_args)
        return result

    async def watch_node_until_condition(
        self, key: str, condition: Callable[[Any], bool], config: SubscriptionConfig
    ):
        """
        Creates a subscription that listens for data change notifications on a node until a condition is met
        Args:
            key: the node to listed
            condition: the condition that should be met to stop the subscription
                       every time asyncua library receives a data change notification
                       it will check if the condition is met
            config: configuration parameters for the subscription.
                    https://reference.opcfoundation.org/Core/Part4/v104/docs/5.13.1
        """
        data_change_sub = DataChangeSubscription(
            condition, print_received_messages=config.print_received_messages
        )

        create_sub_params = CreateSubscriptionParameters(
            RequestedPublishingInterval=config.requested_publishing_interval,
            RequestedLifetimeCount=config.requested_lifetime_count,
            RequestedMaxKeepAliveCount=self._client.get_keepalive_count(
                config.requested_publishing_interval
            ),
            MaxNotificationsPerPublish=config.max_notifications_per_publish,
            PublishingEnabled=True,
            Priority=config.priority,
        )
        sub = await self._client.create_subscription(create_sub_params, data_change_sub)
        self._compare_subscription_parameters(create_sub_params, sub.parameters)

        node = self._client.get_node(key)
        await sub.subscribe_data_change(
            nodes=node, queuesize=config.queue_size, sampling_interval=config.sampling_interval
        )

        await data_change_sub
        await sub.delete()

    async def _map_arguments(self, function_node, *args):
        input_arguments_node = await function_node.get_child("0:InputArguments")
        required_arguments = await input_arguments_node.get_value()
        if len(args) != len(required_arguments):
            raise Exception(  # pylint: disable=broad-exception-raised
                f"you gave: {len(args)} number of argument but the opcua function expects: {len(required_arguments)}"
            )

        mapped_input_argument_values = []
        for i, argument in enumerate(required_arguments):
            data_type_identifier = argument.DataType.Identifier
            mapped_input_argument_values.append(Variant(args[i], VariantType(data_type_identifier)))
        return mapped_input_argument_values

    async def get_node(self, key: str) -> asyncua.Node:
        return self._client.get_node(key)

    def _compare_subscription_parameters(self, create_sub_params, revised_sub_params):
        if (
            create_sub_params.RequestedPublishingInterval
            == revised_sub_params.RequestedPublishingInterval
            and create_sub_params.RequestedLifetimeCount
            == revised_sub_params.RequestedLifetimeCount
            and create_sub_params.RequestedMaxKeepAliveCount
            == revised_sub_params.RequestedMaxKeepAliveCount
        ):
            return

        # this print statement is here to inform Wandelscript user about the differences between the subscription parameters
        logger.warning(
            f"Revised values returned differ from subscription values: {revised_sub_params}"
        )
        print(f"Revised values returned differ from subscription values: {revised_sub_params}")



import asyncio
import logging

import requests
import wandelbots_api_client as wb
from pydantic import BaseModel, Field
from wandelbots_api_client.models.all_joint_positions_request import AllJointPositionsRequest
from wandelbots_api_client.models.all_joint_positions_response import AllJointPositionsResponse

from nova.api import models
from nova_rerun_bridge.benchmark.benchmark_base import BenchmarkStrategy, run_benchmark

logger = logging.getLogger(__name__)


class CollisionFreeP2PRequest(BaseModel):
    """Request model for collision-free point-to-point movement planning."""

    plan: wb.models.PlanCollisionFreePTPRequest = Field(
        ..., description="NOVA plan request for collision-free p2p movement"
    )
    n_control_points: int = Field(
        default=200, description="Number of control points for the trajectory"
    )
    n_eval_points: int = Field(
        default=1000, description="Number of evaluation points for the trajectory"
    )
    n_iterations: int = Field(
        default=10_000, description="Number of iterations for the optimization"
    )
    n_logging_interval: int = Field(default=250, description="Number of iterations between logs")
    collision_margin: int = Field(
        default=10, description="Collision margin between colliders in mm"
    )


class CollisionFreeMagmaStrategy(BenchmarkStrategy):
    """Strategy for collision-free motion planning using Magma P2P."""

    name = "collision_free_magma"

    async def _get_valid_configurations(
        self, nova, motion_group, target, tcp
    ) -> list[AllJointPositionsResponse] | None:
        """Calculate valid inverse kinematic configurations."""
        try:
            response = (
                await nova._api_client.motion_group_kinematic_api.calculate_all_inverse_kinematic(
                    cell=nova.cell().cell_id,
                    motion_group=motion_group.motion_group_id,
                    all_joint_positions_request=AllJointPositionsRequest(
                        motion_group=motion_group.motion_group_id,
                        tcp_pose=models.TcpPose(
                            position=models.Vector3d(
                                x=target.position[0], y=target.position[1], z=target.position[2]
                            ),
                            orientation=models.Vector3d(
                                x=target.orientation[0],
                                y=target.orientation[1],
                                z=target.orientation[2],
                            ),
                            tcp=tcp,
                        ),
                    ),
                )
            )
            return response.joint_positions
        except (wb.ApiException, requests.RequestException, ValueError, Exception) as e:
            logger.exception("Failed to calculate inverse kinematics: %s", e)
            return None

    async def plan(
        self,
        motion_group,
        target,
        collision_scene,
        tcp,
        optimizer_setup,
        nova,
        start_joint_position,
    ):
        """Plan a collision-free trajectory."""
        valid_configs = await self._get_valid_configurations(nova, motion_group, target, tcp)
        if not valid_configs:
            raise ValueError("No valid configurations found")

        collision_motion_group = None
        if collision_scene and collision_scene.colliders:
            if (
                collision_scene.motion_groups
                and optimizer_setup.motion_group_type in collision_scene.motion_groups
            ):
                collision_motion_group = collision_scene.motion_groups[
                    optimizer_setup.motion_group_type
                ]

        headers = {"Accept": "application/json", "Content-Type": "application/json"}

        for config in valid_configs:
            try:
                plan_request = wb.models.PlanCollisionFreePTPRequest(
                    robot_setup=optimizer_setup.model_dump(),
                    start_joint_position=start_joint_position,
                    static_colliders=collision_scene.colliders,
                    collision_motion_group=collision_motion_group,
                    target=wb.models.PlanCollisionFreePTPRequestTarget(config.joints),
                )

                request = CollisionFreeP2PRequest(
                    plan=plan_request,
                    n_control_points=50,
                    n_eval_points=200,
                    n_iterations=2000,
                    n_logging_interval=25,
                    collision_margin=5,
                )

                response = requests.post(
                    f"{nova._api_client._host}/{nova.cell().cell_id}/magma-p2p/planCollisionFreeP2P",
                    headers=headers,
                    json=request.model_dump(mode="json"),
                )

                if response.status_code == 200:
                    return wb.models.PlanTrajectoryResponse.model_validate(
                        response.json()
                    ).response.actual_instance

                logger.error(f"Planning failed with status {response.status_code}: {response.text}")
            except (wb.ApiException, requests.RequestException, ValueError, Exception) as e:
                logger.exception("Error during planning: %s", e)
                continue

        raise RuntimeError("Failed to plan trajectory for all configurations")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    asyncio.run(run_benchmark(CollisionFreeMagmaStrategy()))



from .client import ClientConfig, OPCUAClient
from .functions import opcua_call, opcua_read, opcua_write, wait_for_opcua_value

__all__ = [
    "OPCUAClient",
    "ClientConfig",
    "opcua_read",
    "opcua_write",
    "opcua_call",
    "wait_for_opcua_value",
]



import wandelbots_api_client as wb


def compare_collition_scenes(scene1: wb.models.CollisionScene, scene2: wb.models.CollisionScene):
    if scene1.colliders != scene2.colliders:
        return False

    # Compare motion groups
    if scene1.motion_groups != scene2.motion_groups:
        return False

    return True
